#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <limits.h>
#include "patricia.h"

/******************************************
 *          EXTERNAL VARIABLES            *
 ******************************************/

extern int max_trans_len; 
extern int TMAX; 
extern int min_sup;
extern long long num_node;
extern FILE* out;

/******************************************
 *           GLOBAL  VARIABLES            *
 ******************************************/
long  PAT_size=0;
long  HASH_size=0;
int   freq=0;       
int*  itemset;
long* freq_itemset;
TRANSACTION temp,t; 
int  PARAM1=3;  /* Used to decide dataset projection */ 
int  PARAM2=10; /* Used to decide dataset projection */ 
int  first=true;/*First patricia construction*/
/******************************************
 *                  main                  *
 ******************************************/

int main(int argc,char ** argv){

  IL * ilp;
  pNODE fp,*tree;
  int i;
  int last_transaction_read;
  long num_trans=0; 
  
  /* Read input parameters */
  read_param(argc, argv);   

  /* Open transaction file */
  init_transaction_read ();

  /* Open Output file, if any */
  init_out();

  /*-----------------------------------------------*
   * FIRST PASS ON THE DATASET, AND INITIALIZATION *
   * OF count_element VECTOR and IL                *
   *-----------------------------------------------*/

  init_count();
  while (getNextTransaction (NULL));

  /* Construct IL */
  ilp=make_IL(&freq);
  /*If there aren't frequent itemset print empty-set frequency and exit*/

  /*-----------------------------------------------*
   * SECOND PASS ON THE DATASET, AND CONSTRUCTION  *
   * OF PATRICIA TREE                              *
   *-----------------------------------------------*/

  /* Initialize support variables */
  TMAX=max_trans_len+1; /* Max transaction size+1 */
  if((t=(TRANSACTION)malloc((max_trans_len+2)*sizeof(int)))==NULL) 
   pat_error("Error: transaction allocation"); 
  if((temp=(TRANSACTION)malloc((max_trans_len+2)*sizeof(int)))==NULL)
   pat_error("Error: transaction allocation"); 
  if((itemset=(int *)calloc(freq,sizeof(int)))==NULL)
   pat_error("Error: itemset array allocation"); 
  if((freq_itemset=(long *)calloc(freq,sizeof(long)))==NULL)
   pat_error("Error: freq_itemset array allocation"); 

  /* create root of tree */
  fp=create_node(NULL,NULL); 

  /* read one transaction at the time and insert it in the tree */
  last_transaction_read=0;
  while (!last_transaction_read) {
    if (getNextTransaction(&t)) {
      num_trans++;
      FP_tree_insert(&fp,&t[0]);
    } 
    else {
      if (t[0]!=0){
        num_trans++;
        FP_tree_insert(&fp,&t[0]);
      }
      last_transaction_read = 1;
    }
  }

  /* Initialize array of pointers to nodes */
  if((tree=(pNODE*)calloc(num_node+1,sizeof(pNODE)))==NULL)
   pat_error("Error: tree array allocation");  
  tree[num_node]=NULL;
  num_node=0;

  /* Link Patricia tree to IL*/
  visit_tree(ilp,fp,tree,0,NULL);

  /*Valute the euristic*/
  if(malloc(2*PAT_size)==NULL){
   if(malloc(PAT_size)==NULL)
	   PARAM1=0;/*Disable rebuild*/
   else
	   PARAM1=2;/*Enable only first level's rebuild*/
  }
  first=false;/*No more calculate PAT_size*/
  /*-----------------------------------------------*
   *        FREQUENT ITEMSETS COMPUTATION          *
   *-----------------------------------------------*/
  if(freq==0){
   printf("1 \n"); /* 1 freq. itemset of length 0 */
   if ((out) && num_trans>min_sup) fprintf(out, "(%ld)\n", num_trans);
   exit(0);
  }
  if (out) fprintf(out, "(%ld)\n", num_trans);

  FreqItemCompute(ilp, freq, 0);

  /*-----------------------------------------------*
   *    PRINT #FREQUENT ITEMSETS FOR EACH LENGTH   *
   *-----------------------------------------------*/

  printf("1 \n"); /* 1 freq. itemset of length 0 */
  i=0;
  while(freq_itemset[i]!=0){
    printf("%d\n",freq_itemset[i++]); /* #freq. itemsets of length i+1 */
  }

  /* Close Output file, if any */
  done_write();

  /* Free memory occupied by Patricia tree */
  while(tree[i]!=NULL) free(tree[i++]);
  free(tree); 
  free(fp);

  /* exit */
  return 1;
}
