/**********************************************************************
 *                   ITEMSET COMPUTATION FUNCTIONS                    *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <limits.h>
#include "patricia.h"



/******************************************
 *          EXTERNAL VARIABLES            *
 ******************************************/




extern long* freq_itemset;
extern long  long num_node;
extern int   min_sup;
extern int*  itemset;
extern TRANSACTION temp,t;
extern FILE* out;
extern int  PARAM1;
extern int  PARAM2;
extern long  PAT_size;

/******************************************
 *           GLOBAL  VARIABLES            *
 ******************************************/

int  TMAX=0; /* Max transaction size*/
int  direct; /* Flag to decide invokation of DirectCompute*/
int  current_freq; /* Number of currently frequent items */

/******************************************
 *             DirectCompute              *
 ******************************************/
 
void DirectCompute(IL *ilp, int l, int s){

  int curr, i, num_subsets;

  /* Add all nonempty subsets of temp[1..temp[0]] to itemset[0..s-1] */
  curr=temp[0];
  if (out) {
    while (curr >= 1) {
      if (temp[curr] == itemset[s-1]) {curr--; s--;}
      else {
        freq_itemset[s]++;
        itemset[s++] = temp[curr];
        printSet(s,itemset,ilp[l].count); /* Output itemset[0..s-1] */
        curr=temp[0];
      }
    }
  }
  else {
    num_subsets=curr;
    for (i=1; i<=curr; i++){
      freq_itemset[s++]+=num_subsets;
      num_subsets*=(curr-i);
      num_subsets/=(i+1);
    }
  }
}

/******************************************
 *                 Build                  *
 ******************************************/
 
void  Build(IL *oldilp, IL *newilp, pNODE **newtree, int l) {
 
  pNODE newfp,v,next;
  int i,j,x,tmp,pos, new_ilp_size;
  
  /*-----------------------------------------------*
   *                CREATE NEW IL                  *
   *-----------------------------------------------*/
 
  /* Initialize newilp */
  pos=0;
  for (i=0; i<l; i++){
    if (oldilp[i].count > min_sup) {
      newilp[pos].item_id=i; /* Pointer to oldilp */
      newilp[pos++].count= oldilp[i].count;
    }
    else oldilp[i].count=-1; /* Denotes non frequent item */
  }

  /* Sort newilp by count */
  new_ilp_size = pos;
  IL_sort(newilp,0,new_ilp_size-1); 

  /* Set correct item_id in newilp and pointers from oldilp to newilp */
  for (i=0; i<pos; i++) {
    j=newilp[i].item_id; /* j = pointer to oldilp*/
    oldilp[j].count = i; /* set oldilp[j].count to point to newilp*/
    newilp[i].item_id=oldilp[j].item_id; /* set correct item_id in newilp*/
  }

  /*-----------------------------------------------*
   *           CREATE NEW PATRICIA TREE            *
   *-----------------------------------------------*/

  /* Create root of tree */
  newfp=create_node(NULL,NULL);

  /* 
   * For each node in list oldilp[l].ptr1: walk up the old tree, 
   * and insert transaction found along path into the new tree  
   */
  v =(pNODE)oldilp[l].ptr1;                       
  while (v != NULL) {

    int h=ITEMS;
    int count=v[COUNT].val;
    i = TMAX-1; /* t[TMAX] = last position in t*/
   
    /* Insert in t items up to and excluded l */
    while((j=v[h].val)!=l) {
      tmp=oldilp[j].count; /* pointer to newilp*/
      if (tmp > -1) t[i--] = tmp;
      h++;
    }

    /* Save next node in oldilp[l].ptr1 to be visited */
    next=v[h].pt; 

    /* Walk from v to the root of tree */
    h=v[LAST_EL].pt-v[FATHER].pt; /* Index of last item in v's father */
    v=v[FATHER].pt;
    while(v[FATHER].pt != NULL){
      while(h != ITEMS-1) {
        j=v[h].val; 
        tmp=oldilp[j].count; /* pointer to newilp*/
        if (tmp > -1) t[i--] = tmp;
        h--;
      }
      h=v[LAST_EL].pt-v[FATHER].pt;
      v=v[FATHER].pt;
    }

    /* Set transaction's size and multiplicity */
    t[i] = TMAX-i-1; /* size */
    t[TMAX] = count; /* multiplicity */

    /* Sort transaction in ascending order */
    Tran_sort(t, i+1, TMAX-1);

    /* Insert transaction into new tree */
    FP_tree_insert(&newfp,&t[i]);

    /* Assign to v next node in oldilp[l].ptr1 to be visited */
    v =next;
  } /* end while (v <> NULL) */

  /* Set ptr1 and ptr2 pointers in newilp */
  for(i=0;i<new_ilp_size;i++)
    newilp[i].ptr1=newilp[i].ptr2=NULL;

  /* Create vector of pointers to tree nodes */ 
  if(((*newtree)=(pNODE*)calloc(num_node+1,sizeof(pNODE)))==NULL)
    pat_error("Error: newtree allocation");
  (*newtree)[num_node]=NULL; /* Sentinel */
  num_node=0; /* Index used for newtree */

  /* Link new Patricia tree to newilp */
  visit_tree(newilp,newfp, *newtree,0,NULL);
}


/******************************************
 *           Traverse_and_Count           *
 ******************************************/
 
void Traverse_and_Count(IL *ilp, int l, int timestamp){

  pNODE v,u,tempnode;
  pNODE * templist;
  IL * p;
  int i,j,x=1,count;
  int val=ilp[l].count;
  int k=1;
  int curr=l; /* Starting position in  IL */
  int lcount =ilp[curr].count;
  current_freq=0;
  direct=true;
  temp[0]=0;

  p=&ilp[curr];

  /* If ilp[curr].ptr1 was NULL then ilp[curr].ptr2 !=NULL */
  if(ilp[curr].ptr1==NULL)  v=(pNODE)ilp[curr].ptr2;
  else v=(pNODE)ilp[curr].ptr1;

  /* 
   * If ilp[curr].ptr1 consists of a single node, 
   * prepare to invoke DirectCompute 
   */
 
  i=ITEMS;
  while((j=v[i].val)!=curr) i++;
  if (v[i].pt==NULL) {
    int num_elem=v[LAST_EL].pt-((pNODE)&((v[FATHER].pt)[ITEMS]));
    for(j=i-1;j>=ITEMS;j--) temp[i-j]=ilp[v[j].val].item_id;
    /* Walk up the tree */
    v=v[FATHER].pt;
    j=i-j;
    while(v[FATHER].pt != NULL){
      for(i=num_elem+ITEMS;i>=ITEMS;i--)
        temp[j++] = ilp[v[i].val].item_id;
      num_elem=v[LAST_EL].pt-((pNODE)&((v[FATHER].pt)[ITEMS]));
      v=v[FATHER].pt;
    }
    temp[0]=j-1;
    return;
  }

  /* Reset ilp fields up to entry curr-1 */
  for (i=0; i<curr; i++) {
    ilp[i].count=0;
    ilp[i].ptr1=ilp[i].ptr2=NULL;
  }

  /* 
   * Walk up the subtree whose leaves are the nodes in ilp[curr].ptr1: 
   * visited nodes are inserted in ptr2 lists; nodes still to be 
   * visited are inserted in ptr1 lists
   */
  
  while(true){ 
    i=ITEMS;
    count = v[COUNT].val;
    while((j=v[i].val)!=curr){
      ilp[j].count += count; 
      templist = (pNODE *)&(ilp[j].ptr2);
      tempnode = (*templist);
      (*templist) = v;
      v[i].pt = tempnode;
      i++;
    }
    u = v[FATHER].pt;
    if (u[FATHER].pt != NULL){
      j = (v[LAST_EL].pt)->val; /* last item in u */
      ilp[j].count += count;
      if (u[T_STAMP].val == timestamp) u[COUNT].val+=count;
      else {
        u[T_STAMP].val=timestamp;
        u[COUNT].val=count;
        /* insert u in list ilp[j].ptr1 */
        templist = (pNODE *)&(ilp[j].ptr1);
        tempnode=(*templist);
        (*templist) = u;
        v[LAST_EL].pt->pt = tempnode;
      }
    } /* end if (u[FATHER].pt != NULL) */
    if(v[i].pt!=NULL) v = v[i].pt;
    else break;
    }   
    /*End of ilp[curr].ptr1 list*/
    while(true){
    /* Attach list ilp[curr].ptr2 to end of list ilp[curr].ptr1 */
    v[i].pt=(pNODE)((*p).ptr2);
    /* Find new list for continuing walking up the tree */
    p--;
    while((*p).ptr1==NULL){
    if ((*p).count > min_sup){
     current_freq++; 
     if(direct){
      if((*p).count == val){
       temp[x++] = (*p).item_id;
      }
      else{
       direct=false;
      }
     }
    }
    p--;
    }/*End while((*p).ptr1==NULL)*/
    if (p < ilp) break;
    else{
     if ((*p).count > min_sup){
     current_freq++; 
     if(direct){
      if((*p).count == val){
       temp[x++] = (*p).item_id;
      }
      else{
       direct=false;
      }
     }
     }
     v=(pNODE)((*p).ptr1);
     while(true){
     i=ITEMS;
     count = v[COUNT].val;
     while(v[i+1].val!=-1){
      j=v[i].val;
      ilp[j].count += count; 
      templist = (pNODE *)&(ilp[j].ptr2);
      tempnode = (*templist);
      (*templist) = v;
      v[i].pt = tempnode;
      i++;
     }
     u = v[FATHER].pt;
     if (u[FATHER].pt != NULL){
      j = (v[LAST_EL].pt)->val; /* last item in u */
      ilp[j].count += count;
      if (u[T_STAMP].val == timestamp) u[COUNT].val+=count;
      else {
       u[T_STAMP].val=timestamp;
       u[COUNT].val=count;
       /* insert u in list ilp[j].ptr1 */
       templist = (pNODE *)&(ilp[j].ptr1);
       tempnode=(*templist);
       (*templist) = u;
       v[LAST_EL].pt->pt = tempnode;
      }
     } /* end if (u[FATHER].pt != NULL) */
     if(v[i].pt!=NULL) v = v[i].pt;
     else break;
     }
    }/* end else (if p<ilp )*/
 } /*end while (true)*/
 /*End of il*/
 temp[0]=x-1; /* Set the number of elements of temp (for DirectCompute)*/
}




  





/******************************************
 *            FreqItemCompute             *
 ******************************************/
 
void FreqItemCompute(IL *ilp, int nofreq, int s) {

  int i;
  int l=0;
  int rebuildflag=false;
  int timestamp=0; 
  IL *newilp; 
  pNODE *newtree; 

  /* Iterative depth-first exploration of Frequent ItemSet Tree (FIST) */
  while (l < nofreq) {
    if (ilp[l].count <= min_sup) l++;
    else {
      if ((s>0) && (ilp[l].item_id == itemset[s-1])) {l++; s--;}
      else { /* Found new node of FIST */

        /* Produce associated itmeset */ 
        freq_itemset[s]++;
	itemset[s++] = ilp[l].item_id;
        if(out) printSet(s,itemset,ilp[l].count); /* Output itemset[0..s-1] */

        /* Prepare for visiting children */
	timestamp++;
        if ((timestamp+nofreq)==(INT_MAX-1)) rebuildflag=true; 
        Traverse_and_Count(ilp, l, timestamp);

        /* If all children have same frequency of father, call DirectCompute */
        if (direct){
           DirectCompute(ilp,l,s);
           s--;
           l++;
        }
        else {
          /* If appropriate, project dataset */
          if (rebuildflag || ((s < PARAM1) && (current_freq > PARAM2))) {
         
            /* Allocate new IL */
            if((newilp = (IL*)malloc((1+current_freq)*sizeof(struct itemset_list)))==NULL)
	      pat_error("error: IL allocation");
            /*Set the terminator of il*/
            newilp[0].ptr1=(void *)&newilp[1];
            newilp=&newilp[1];
           
            num_node=0;

 	    /* Initialize new IL and create new Patricia tree */
            Build(ilp, newilp, &newtree, l);

            /* Recurse */      
            FreqItemCompute(newilp, current_freq, s);

            /* Free space taken by new tree and new IL */
	     i=0;
             while(newtree[i]!=NULL) free(newtree[i++]);
	     newilp=&newilp[-1];/*necessary for freeing*/
             free(newilp);
	     free(newtree);
	     s--;
             l++;
          }
          else l=0; /* No projection ==> go to beginning of IL */
	} /* end if (direct) */
      }/* end if ((s>0) && (ilp[l].item_id == itemset[s-1])) */
    } /* end if (ilp[l].count <= min_sup) */
  } /* end while (l < nofreq) */
} 

