/**********************************************************************
 *                    TREE CONSTRUCTION FUNCTIONS                     *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <limits.h>
#include "patricia.h"


/******************************************
 *          EXTERNAL VARIABLES            *
 ******************************************/

extern long  max_item;
extern int*  count_element;
extern int   min_sup; 
extern TRANSACTION temp; 


/******************************************
 *           GLOBAL  VARIABLES            *
 ******************************************/

int  hash_limit=7;
long long num_node=0;

/******************************************
 *               Tran_sort                *
 ******************************************/

void Tran_sort(TRANSACTION t, int left, int right) {
 
  int pivot;
  int p, l, r, temp;
   
  if (left >= right) return;
  p=(left+right)/2; /* pivot index */
  pivot = t[p]; /* pivot value */
 
  /* swap pivot with last element */
  t[p] = t[right];
  t[right] = pivot; 
  /* Partition elements between left and right-1 */
  l=left; r=right-1;
  while (l<=r) {
    while ((l<=r) && (t[l] <= pivot)) l++;
    while ((r>=l) && (t[r] >= pivot)) r--;
    if (l<r) {
      temp = t[l];
      t[l] = t[r];
      t[r] = temp;
    }
  }
  temp = t[l];
  t[l] = t[right];
  t[right] = temp;
 
  /* Recurse */
  Tran_sort(t, left, l-1);
  Tran_sort(t, l+1,right);
}
 
/******************************************
 *               IL_sort                *
 ******************************************/
 
void IL_sort(IL *ilp, int left, int right) {

  int pivot;            
  int p, l, r;
  IL temp;       
  
  if (left >= right) return;
  p=(left+right)/2; /* pivot index */
  pivot = ilp[p].count; /* pivot value */

  /* swap pivot with last element */
  temp.item_id = ilp[p].item_id;
  temp.count = ilp[p].count;
  ilp[p].item_id = ilp[right].item_id;
  ilp[p].count = ilp[right].count;
  ilp[right].item_id=temp.item_id;
  ilp[right].count=temp.count;

  /* Partition elements between left and right-1 */
  l=left; r=right-1;
  while (l<=r) {
    while ((l<=r) && (ilp[l].count >= pivot)) l++; 
    while ((r>=l) && (ilp[r].count <= pivot)) r--; 
    if (l<r) {
       temp.item_id = ilp[l].item_id;
       temp.count = ilp[l].count;
       ilp[l].item_id = ilp[r].item_id;
       ilp[l].count = ilp[r].count;
       ilp[r].item_id=temp.item_id;
       ilp[r].count=temp.count;
    }
  }
  temp.item_id = ilp[l].item_id;
  temp.count = ilp[l].count;
  ilp[l].item_id = ilp[right].item_id;
  ilp[l].count = ilp[right].count;
  ilp[right].item_id=temp.item_id;
  ilp[right].count=temp.count;

  /* Recurse */
  IL_sort(ilp, left, l-1);
  IL_sort(ilp, l+1,right);
}       

/******************************************
 *               make_IL                  *
 ******************************************/
 
IL *make_IL(int *freq){
  
  int i,j=0;
  IL * ilp;
  
  /* Compute number of frequent items */
  (*freq)=0;
  for(i=0;i<max_item;i++) 
    if(count_element[i]>min_sup) (*freq)++;

  /* Allocate memory for IL */  
  if((ilp=(IL *)malloc(sizeof(struct itemset_list)*((*freq)+1)))==NULL) 
    pat_error("Error: IL allocation");
  ilp[0].ptr1=(void *)&ilp[1];
  ilp=&ilp[1];
  
  /* Initialize IL */  
  for(i=0;i<max_item;i++){
    if(count_element[i]>min_sup){
      ilp[j].item_id=i;
      ilp[j].ptr1=ilp[j].ptr2=NULL;
      ilp[j++].count=count_element[i];
    }
    else{
      count_element[i]=-1;
    }
  }
  
  /* Sort IL by count value */
  IL_sort(ilp,0,(*freq)-1);

  /* Link count_element to IL */    
  for(j=0;j<(*freq);j++) count_element[ilp[j].item_id]=j;

  /* Return pointer to IL */
  return ilp;
}

/******************************************
 *                 f_hash                 *
 ******************************************/
 
int f_hash(int item, int elm_hash){
  return (item % elm_hash);
}
 
/******************************************
 *           insert_in_hash               *
 ******************************************/
 
void insert_in_hash(hash *fg, pNODE *f, int pos){

  pNODE *c, tmp;
  int elem;
  int new_elem;
 
  c=&((*fg)[pos]); /* Bucket for insertion */
  new_elem=(*f)[ITEMS].val;
  
  /* Find correct position in bucket */
  while ((*c)!=NULL){
    elem=(*c)[ITEMS].val;
    if(elem > new_elem) break;
    c=&((*c)[rSIBLI].pt);
  }

  if((*c)==NULL){ /* insert at the end of bucket */
    (*c)=(*f);
    (*c)[rSIBLI].pt=NULL;
  }
  else{ 
    tmp=(*c);
    (*c)=(*f);
    (*c)[rSIBLI].pt=tmp;
  }
}

/******************************************
 *                new_hash                *
 ******************************************/
 
void new_hash(pNODE *p, int num_buckets){

  hash fg;
  pNODE f,next;
  int i=0;

  /* Create new hash table */
  if((fg=(pNODE *)calloc(num_buckets,sizeof(hash)))==NULL)
   pat_error("Error: new hash allocation");
  
  /* Insert node's children into new hash table */
  for(i=0;i<num_buckets/2;i++){
    f=((hash)((*p)[CHILDREN].pt))[i];
    while (f!=NULL){
      next=f[rSIBLI].pt;
      insert_in_hash(&fg,&f,f_hash(f[ITEMS].val,num_buckets));
      f=next;
    }
  }

  /* Free old hash table and assign new one to node */
  free((*p)[CHILDREN].pt);
  (*p)[CHILDREN].pt=(pNODE)fg;
  (*p)[ITEMS].pt=(pNODE)(&(((hash)(*p)[CHILDREN].pt)[num_buckets-1]));
}

/******************************************
 *              create_node               *
 ******************************************/
 
pNODE create_node(TRANSACTION t, pNODE father){

  pNODE p;
  int i=0;

  /* create node */
  if(t!=NULL){ /* non-root node */
    if((p=(pNODE)malloc(((t[0]+DIM_NODE)*sizeof(struct NODE))))==NULL){
      pat_error("error: non-root node allocation");
    }
    num_node++;
    p[COUNT].val=t[t[0]+1]; /* t[t[0]+1] = transaction multiplicity */
  }
  else{ /* root node */
    if((p=(pNODE)malloc(((1+DIM_NODE)*sizeof(struct NODE))))==NULL){
      pat_error("error: root allocation");
    }
    p[COUNT].val=1;
  }
  p[rSIBLI].pt=NULL;
  p[CHILDREN].pt=NULL;
  p[NUM_CHILDREN].val=0;
  
  /* insert items (i.e., indices in IL) in node  */ 
  if(father!=NULL){
     for(i=0;i<t[0];i++){
       p[ITEMS+i].val=t[i+1]; /*item_id*/
       p[ITEMS+i].pt=NULL; /* NEXT */
     }
  }

  /* Set sentinel */
  p[ITEMS+i].val=-1;
  p[ITEMS+i].pt=NULL; 
  return p;
}

/******************************************
 *               split_node               *
 ******************************************/
 
void split_node(pNODE *node, int ind1){

  int i;
  pNODE *c;
  hash fg;

  /* Save pointer to old hash table for node */
  fg=(hash)((*node)[CHILDREN].pt); 

  /* Create transaction with last segment of items */
  i=0;
  while((*node)[ITEMS+ind1+i].val != -1){
    temp[i+1]=(*node)[ITEMS+ind1+i].val;
    i++;
  }
  temp[0]=i;
  temp[temp[0]+1]=(*node)[COUNT].val;

  /* Resize node */
  if(((*node)=(pNODE)realloc((*node),(DIM_NODE+ind1)*sizeof(struct NODE)))==NULL)
   pat_error("Error: resize node in split node");
  (*node)[ITEMS+ind1].val=-1;

  /* Create new hash table of size 1 for node */
  if(((*node)[CHILDREN].pt=(pNODE)calloc(1,sizeof(hash)))==NULL)
   pat_error("Error:  new hash table of size 1 for node in split node");
  c=&(((hash)(*node)[CHILDREN].pt)[0]); /* unique bucket in new tale */
  
  /* Insert transaction temp as (unique) child of node */
  (*c)=create_node(temp,*node);
  (*c)[NUM_CHILDREN].val=(*node)[NUM_CHILDREN].val;
  (*node)[NUM_CHILDREN].val=1;
  (*c)[ITEMS].pt=(*node)[ITEMS].pt; /* Points to end of hash table */
  (*c)[CHILDREN].pt=(pNODE)fg; /* Points to beginning of hash table */
  (*node)[ITEMS].pt=(*node)[CHILDREN].pt;
}

/******************************************
 *              split2_node               *
 ******************************************/
 
void  split_2node(pNODE *node, int ind1, TRANSACTION tran){

  int i;
  pNODE *c,tp;
  hash fg;

  /* Save pointer to old hash table for node */
  fg=(hash)((*node)[CHILDREN].pt);

  /* Create transaction with last segment of items */
  i=0;
  while((*node)[ITEMS+ind1+i].val != -1){
    temp[i+1]=(*node)[ITEMS+ind1+i].val;
    i++;
  }
  temp[0]=i;
  temp[temp[0]+1]=(*node)[COUNT].val;

  /* Resize node */
  if(((*node)=(pNODE)realloc((*node),(DIM_NODE+ind1)*sizeof(struct NODE)))==NULL)
   pat_error("Error: resize node in split node");
  (*node)[ITEMS+ind1].val=-1;

  /* Create new hash table of size 1 for node */
  if(((*node)[CHILDREN].pt=(pNODE)calloc(1,sizeof(hash)))==NULL)
     pat_error("Error:  new hash table of size 1 for node in split node");

  c=&(((hash)(*node)[CHILDREN].pt)[0]); /* unique bucket in new tale */
  
  /* Insert transaction temp as child of node */
  (*c)=create_node(temp,*node);
  (*c)[NUM_CHILDREN].val=(*node)[NUM_CHILDREN].val;
  (*node)[NUM_CHILDREN].val=2;
  (*c)[ITEMS].pt=(pNODE)(*node)[ITEMS].pt; /* Points to end of hash table */
  (*c)[CHILDREN].pt=(pNODE)fg; /* Points to beginning of hash table */
  (*node)[ITEMS].pt=(*node)[CHILDREN].pt;

  /* Insert transaction tran as child of node */
  if ((*c)[ITEMS].val > tran[1] ){
    tp=(*c);
    (*c)=create_node(tran,*node);
    (*c)[rSIBLI].pt=tp;
  }
  else{
    tp=create_node(tran,*node);
    (*c)[rSIBLI].pt=tp;
  }
}

/******************************************
 *            FP_tree_insert              *
 ******************************************/
 
void FP_tree_insert(pNODE *p, TRANSACTION t){

  pNODE *c,tp;
  int i;
  int elem=0;
  int num_bucket=0;
  
  if (t[0]==0) return; /* t[0]=transaction size */
  
  /* Compute hash table size for node (*p) */
  if((*p)[ITEMS].pt == NULL) 
    num_bucket=1;
  else
    num_bucket=1+(int)(((hash)(*p)[ITEMS].pt)-&(((hash)(*p)[CHILDREN].pt)[0])); 
  /* If hash table does not exist, create one */
  if((*p)[CHILDREN].pt==NULL){
    if(((*p)[CHILDREN].pt=(pNODE)calloc(1,sizeof(hash)))==NULL)
        pat_error("Error: Hash Allocation in FP_tree_insert");
    (*p)[ITEMS].pt=(*p)[CHILDREN].pt;        
   }

  /* Compute bucket where to look for correct child */
  c=&(((hash)(*p)[CHILDREN].pt)[f_hash(t[1],num_bucket)]);
  while((*c)!=NULL){
    elem=(*c)[ITEMS].val;
    if ( elem >= t[1] ) break;
    else c=&((*c)[rSIBLI].pt);
  }
  if ((*c)==NULL){ 
    /* Create new child */
    (*p)[NUM_CHILDREN].val +=1;
    (*c)=create_node(t,*p);
  }
  else{
    /* Continue along found child */
    if (elem==t[1]){
      int c_t=t[t[0]+1]; /* transaction multiplicity */
      i=1;
      while((*c)[ITEMS+i].val != -1){

        /* Case 1: t contained in (*c) */
        if(i+1>t[0]){
          split_node(&(*c),i);
          ((*c)[COUNT].val) += c_t;
	  i=t[0]+1;
          break;
        }

        /* Case 2: partial match between t and (*c) */
        if((*c)[ITEMS+i].val != t[i+1]){
          t[i]=t[0]-i;
          t[i+t[i]+1]=c_t;
          split_2node(&(*c),i,&t[i]);
          ((*c)[COUNT].val)+= c_t;
          i=t[0]+1;
          break;
        }
        i++;
      } /* end while ((*c)[ITEMS+i].val != -1) */
      
      /* Case 3: t contains (*c) */
      if(i<=t[0]){
        ((*c)[COUNT].val)+= c_t;
        if(i<t[0]){
          t[i]=t[0]-i;
          t[i+t[i]+1]=t[t[0]+1]; /* Set multiplicity */
          FP_tree_insert(&(*c),&t[i]); /* Recurse */
        }
      }
    
    }
    else{ 
      /* (elem>t[1]) ==> Create new child */
      (*p)[NUM_CHILDREN].val +=1;
      tp=(*c);
      (*c)=create_node(t,*p);
      (*c)[rSIBLI].pt=tp;
    } /* end if (elem==t[1]) */
  }/* end if ((*c)==NULL) */

  /* If hash table is full do rehash */  
  if ((((*p)[NUM_CHILDREN].val)/num_bucket) > hash_limit)
    new_hash(&(*p),(2*num_bucket));
}

/******************************************
 *                 insert                 *
 ******************************************/
 
void insert(pNODE *pl, pNODE newl, int h){
  pNODE p;

  /* Insert node newl at the head of list pl */
  if ((*pl) == NULL) {
    (*pl) = (newl);
    (*pl)[h].pt=NULL;
  } 
  else{
    p=(*pl);
    (*pl) = newl;
    (*pl)[h].pt = p;
  }
}

/******************************************
 *              visit_tree                *
 ******************************************/
 
void visit_tree(IL *ilp, pNODE p, pNODE *newtree, int liv, pNODE pd){

  int i;
  pNODE f,last,tmp,tempnode;
  pNODE * templist;
  int num_bucket;

  /* Compute hash table size for node p */  
  if(p[ITEMS].pt== NULL) 
    num_bucket=1;
  else
    num_bucket=1+(int)(((hash)p[ITEMS].pt)-((hash)p[CHILDREN].pt));      
  
  
  
  /* If non-root node, visit node */
  if (liv!=0) {
    int j;
    int h=ITEMS;
    newtree[num_node++]=p; /* Save pointer to node */
    /* For each item set link from IL */
    while ((j=p[h].val )!= -1){
      /*insert((pNODE *)&(ilp[p[h].val].ptr1),p,h);*/
      templist=(pNODE *)&(ilp[j].ptr1);
      tempnode=(*templist);
      (*templist)=p;
      p[h].pt=tempnode;
      h+=1;
    }
    /*Compute the PATRICIA's occupation if first construction*/
    
    last=&p[h-1]; /* Address of last item */
  }
  else{ 
    last=&p[ITEMS];
    
    
  }
  /* Visit each child recursively and set pointer to last item of father */ 
  for(i=0;i<num_bucket;i++){
    if (p[CHILDREN].pt==NULL) break;
    f=((hash)(p[CHILDREN].pt))[i];
    while (f!=NULL){
      visit_tree(ilp,f,newtree,liv+1,p);
      tmp=f;
      f=f[rSIBLI].pt;
      tmp[LAST_EL].pt=last; /* Pointer to last item of father */
    }
  }

  /* Free hash table in p, if any */
  if (p[CHILDREN].pt != NULL) free(p[CHILDREN].pt);

  /* Set pointer to father */
  p[FATHER].pt=pd;
  p[T_STAMP].val=0;
  return;
}
